/*
 * COPYRIGHT:        See COPYRIGHT.TXT
 * PROJECT:          Ext2 File System Driver for WinNT/2K/XP
 * FILE:             init.c
 * PROGRAMMER:       Matt Wu <mattwu@163.com>
 * HOMEPAGE:         http://ext2.yeah.net
 * UPDATE HISTORY: 
 */

/* INCLUDES *****************************************************************/

#include "ext2fs.h"

/* GLOBALS ***************************************************************/

PEXT2_GLOBAL    Ext2Global = NULL;

/* DEFINITIONS ***********************************************************/

NTSTATUS
DriverEntry(
    IN PDRIVER_OBJECT DriverObject,
    IN PUNICODE_STRING RegistryPath   );

#ifdef ALLOC_PRAGMA
#pragma alloc_text(INIT, Ext2QueryParameters)
#pragma alloc_text(INIT, DriverEntry)
#if EXT2_UNLOAD
#pragma alloc_text(PAGE, DriverUnload)
#endif
#endif

/* FUNCTIONS ***************************************************************/

#if EXT2_UNLOAD

/*
 * FUNCTION: Called by the system to unload the driver
 * ARGUMENTS:
 *           DriverObject = object describing this driver
 * RETURNS:  None
 */

VOID
DriverUnload (IN PDRIVER_OBJECT DriverObject)
{
    UNICODE_STRING  DosDeviceName;

    Ext2Print((DBG_FUNC, "Ext2Fsd: Unloading routine.\n"));

    RtlInitUnicodeString(&DosDeviceName, DOS_DEVICE_NAME);
    IoDeleteSymbolicLink(&DosDeviceName);

    ExDeleteResourceLite(&Ext2Global->LAResource);
    ExDeleteResourceLite(&Ext2Global->CountResource);
    ExDeleteResourceLite(&Ext2Global->Resource);

    ExDeletePagedLookasideList(&(Ext2Global->Ext2McbLookasideList));
    ExDeleteNPagedLookasideList(&(Ext2Global->Ext2CcbLookasideList));
    ExDeleteNPagedLookasideList(&(Ext2Global->Ext2FcbLookasideList));
    ExDeleteNPagedLookasideList(&(Ext2Global->Ext2IrpContextLookasideList));

    Ext2UnloadAllNls();

    IoDeleteDevice(Ext2Global->DeviceObject);
}

#endif

BOOLEAN
Ext2QueryParameters( IN PUNICODE_STRING  RegistryPath)
{
    NTSTATUS                    Status;
    UNICODE_STRING              ParameterPath;
    RTL_QUERY_REGISTRY_TABLE    QueryTable[2];

    ULONG                       WritingSupport;
    ULONG                       CheckingBitmap;
    ULONG                       Ext3ForceWriting;

    UNICODE_STRING              CodePage;
    ANSI_STRING                 AnsiName;

    ParameterPath.Length = 0;

    ParameterPath.MaximumLength =
        RegistryPath->Length + sizeof(PARAMETERS_KEY) + sizeof(WCHAR);

    ParameterPath.Buffer =
        (PWSTR) ExAllocatePool(PagedPool, ParameterPath.MaximumLength);

    if (!ParameterPath.Buffer)
    {
        return FALSE;
    }

    WritingSupport = 0;
    CheckingBitmap = 0;
    Ext3ForceWriting = 0;

    RtlCopyUnicodeString(&ParameterPath, RegistryPath);

    RtlAppendUnicodeToString(&ParameterPath, PARAMETERS_KEY);

    RtlZeroMemory(&QueryTable[0], sizeof(RTL_QUERY_REGISTRY_TABLE) * 2);

    QueryTable[0].Flags = RTL_QUERY_REGISTRY_DIRECT | RTL_QUERY_REGISTRY_REQUIRED;
    QueryTable[0].Name = WRITING_SUPPORT;
    QueryTable[0].EntryContext = &WritingSupport;

    Status = RtlQueryRegistryValues(
        RTL_REGISTRY_ABSOLUTE,
        ParameterPath.Buffer,
        &QueryTable[0],
        NULL,
        NULL        );

    Ext2Print((DBG_USER, "Ext2QueryParameters: WritingSupport=%xh\n", WritingSupport));

    RtlZeroMemory(&QueryTable[0], sizeof(RTL_QUERY_REGISTRY_TABLE) * 2);

    QueryTable[0].Flags = RTL_QUERY_REGISTRY_DIRECT | RTL_QUERY_REGISTRY_REQUIRED;
    QueryTable[0].Name = CHECKING_BITMAP;
    QueryTable[0].EntryContext = &CheckingBitmap;

    Status = RtlQueryRegistryValues(
        RTL_REGISTRY_ABSOLUTE,
        ParameterPath.Buffer,
        &QueryTable[0],
        NULL,
        NULL        );

    Ext2Print((DBG_USER, "Ext2QueryParameters: CheckingBitmap=%xh\n", CheckingBitmap));

    RtlZeroMemory(&QueryTable[0], sizeof(RTL_QUERY_REGISTRY_TABLE) * 2);

    QueryTable[0].Flags = RTL_QUERY_REGISTRY_DIRECT | RTL_QUERY_REGISTRY_REQUIRED;
    QueryTable[0].Name = EXT3_FORCEWRITING;
    QueryTable[0].EntryContext = &Ext3ForceWriting;

    Status = RtlQueryRegistryValues(
        RTL_REGISTRY_ABSOLUTE,
        ParameterPath.Buffer,
        &QueryTable[0],
        NULL,
        NULL        );

    Ext2Print((DBG_USER, "Ext2QueryParameters: Ext3ForceWriting=%xh\n", Ext3ForceWriting));

    RtlZeroMemory(&QueryTable[0], sizeof(RTL_QUERY_REGISTRY_TABLE) * 2);

    QueryTable[0].Flags = RTL_QUERY_REGISTRY_DIRECT | RTL_QUERY_REGISTRY_REQUIRED;
    QueryTable[0].Name = EXT3_CODEPAGE;
    QueryTable[0].EntryContext = &(CodePage);
    CodePage.MaximumLength = CODEPAGE_MAXLEN * sizeof(WCHAR);
    CodePage.Length = 0;
    CodePage.Buffer = &(Ext2Global->CodePage.UniName[0]);

    Status = RtlQueryRegistryValues(
        RTL_REGISTRY_ABSOLUTE,
        ParameterPath.Buffer,
        &QueryTable[0],
        NULL,
        NULL        );
    if (NT_SUCCESS(Status)) {
        Ext2Print((DBG_USER, "Ext2QueryParameters: Ext2CodePage=%S\n", CodePage.Buffer));
        AnsiName.MaximumLength = CODEPAGE_MAXLEN;
        AnsiName.Length = 0;
        AnsiName.Buffer = &(Ext2Global->CodePage.AnsiName[0]);

        Status = RtlUnicodeStringToAnsiString(
                    &AnsiName,
                    &CodePage,
                    FALSE);

        if (!NT_SUCCESS(Status)) {
            Ext2Print((DBG_USER, "Ext2QueryParameters: Wrong CodePage ...\n"));
            RtlCopyMemory(&(Ext2Global->CodePage.AnsiName[0]),"default\0", 8);
        }
    } else {
        Ext2Print((DBG_USER, "Ext2QueryParameters: CodePage not specified.\n"));
        RtlCopyMemory(&(Ext2Global->CodePage.AnsiName[0]),"default\0", 8);
    }

    {
        if (WritingSupport)
        {
            SetFlag(Ext2Global->Flags, EXT2_SUPPORT_WRITING);
        }
        else
        {
            ClearFlag(Ext2Global->Flags, EXT2_SUPPORT_WRITING);
        }

        if (CheckingBitmap)
        {
            SetFlag(Ext2Global->Flags, EXT2_CHECKING_BITMAP);
        }
        else
        {
            ClearFlag(Ext2Global->Flags, EXT2_CHECKING_BITMAP);
        }

        if (Ext3ForceWriting)
        {
            DbgPrint("Ext2Fsd -- Warning: Ext3ForceWriring enabled !!!\n");

            SetFlag(Ext2Global->Flags, EXT3_FORCE_WRITING);
            SetFlag(Ext2Global->Flags, EXT2_SUPPORT_WRITING);
        }
        else
        {
            ClearFlag(Ext2Global->Flags, EXT3_FORCE_WRITING);
        }
    }

    ExFreePool(ParameterPath.Buffer);

    return TRUE;
}


#define NLS_OEM_LEAD_BYTE_INFO            (*NlsOemLeadByteInfo)

#define FsRtlIsLeadDbcsCharacter(DBCS_CHAR) (                      \
    (BOOLEAN)((UCHAR)(DBCS_CHAR) < 0x80 ? FALSE :                  \
              (NLS_MB_CODE_PAGE_TAG &&                             \
               (NLS_OEM_LEAD_BYTE_INFO[(UCHAR)(DBCS_CHAR)] != 0))) \
)


/*
 * NAME: DriverEntry
 * FUNCTION: Called by the system to initalize the driver
 *
 * ARGUMENTS:
 *           DriverObject = object describing this driver
 *           RegistryPath = path to our configuration entries
 * RETURNS: Success or failure
 */
NTSTATUS
DriverEntry (
         IN PDRIVER_OBJECT   DriverObject,
         IN PUNICODE_STRING  RegistryPath
         )
{
    PDEVICE_OBJECT              DeviceObject;
    PFAST_IO_DISPATCH           FastIoDispatch;
    PCACHE_MANAGER_CALLBACKS    CacheManagerCallbacks;
    PEXT2FS_EXT                 DeviceExt;

    UNICODE_STRING              DeviceName;
    NTSTATUS                    Status;

#if EXT2_UNLOAD
    UNICODE_STRING              DosDeviceName;
#endif

    DbgPrint(
        "Ext2Fsd --"
        " Version " 
        EXT2FSD_VERSION
#if EXT2_READ_ONLY
        " (ReadOnly)"
#endif
#if DBG
        " Checked"
#else
        " Free" 
#endif
        " - Built at "
        __DATE__" "
        __TIME__".\n");

    Ext2Print((DBG_FUNC, "Ext2 DriverEntry ...\n"));

    RtlInitUnicodeString(&DeviceName, DEVICE_NAME);
    
    Status = IoCreateDevice(
        DriverObject,
        sizeof(EXT2FS_EXT),
        &DeviceName,
        FILE_DEVICE_DISK_FILE_SYSTEM,
        0,
        FALSE,
        &DeviceObject );
    
    if (!NT_SUCCESS(Status))
    {
        Ext2Print((DBG_ERROR, "IoCreateDevice fs object error.\n"));
        return Status;
    }

    DeviceExt = (PEXT2FS_EXT) DeviceObject->DeviceExtension;
    RtlZeroMemory(DeviceExt, sizeof(EXT2FS_EXT));
    
    Ext2Global = &(DeviceExt->Ext2Global);

    Ext2Global->Identifier.Type = EXT2FGD;
    Ext2Global->Identifier.Size = sizeof(EXT2_GLOBAL);
    Ext2Global->DeviceObject = DeviceObject;
    Ext2Global->DriverObject = DriverObject;

    Ext2QueryParameters(RegistryPath);

    DriverObject->MajorFunction[IRP_MJ_CREATE]              = Ext2BuildRequest;
    DriverObject->MajorFunction[IRP_MJ_CLOSE]               = Ext2BuildRequest;
    DriverObject->MajorFunction[IRP_MJ_READ]                = Ext2BuildRequest;
    DriverObject->MajorFunction[IRP_MJ_WRITE]               = Ext2BuildRequest;

    DriverObject->MajorFunction[IRP_MJ_FLUSH_BUFFERS]       = Ext2BuildRequest;
    DriverObject->MajorFunction[IRP_MJ_SHUTDOWN]            = Ext2BuildRequest;

    DriverObject->MajorFunction[IRP_MJ_QUERY_INFORMATION]   = Ext2BuildRequest;
    DriverObject->MajorFunction[IRP_MJ_SET_INFORMATION]     = Ext2BuildRequest;

    DriverObject->MajorFunction[IRP_MJ_QUERY_VOLUME_INFORMATION]    = Ext2BuildRequest;
    DriverObject->MajorFunction[IRP_MJ_SET_VOLUME_INFORMATION]      = Ext2BuildRequest;

    DriverObject->MajorFunction[IRP_MJ_DIRECTORY_CONTROL]   = Ext2BuildRequest;
    DriverObject->MajorFunction[IRP_MJ_FILE_SYSTEM_CONTROL] = Ext2BuildRequest;
    DriverObject->MajorFunction[IRP_MJ_DEVICE_CONTROL]      = Ext2BuildRequest;
    DriverObject->MajorFunction[IRP_MJ_LOCK_CONTROL]        = Ext2BuildRequest;

    DriverObject->MajorFunction[IRP_MJ_CLEANUP]             = Ext2BuildRequest;

#if (_WIN32_WINNT >= 0x0500)
    DriverObject->MajorFunction[IRP_MJ_PNP]                 = Ext2BuildRequest;
#endif //(_WIN32_WINNT >= 0x0500)

#if EXT2_UNLOAD
    DriverObject->DriverUnload                              = DriverUnload;
#else
    DriverObject->DriverUnload                              = NULL;
#endif

    //
    // Initialize the fast I/O entry points
    //
    
    FastIoDispatch = &(Ext2Global->FastIoDispatch);
    
    FastIoDispatch->SizeOfFastIoDispatch        = sizeof(FAST_IO_DISPATCH);
    FastIoDispatch->FastIoCheckIfPossible       = Ext2FastIoCheckIfPossible;
#if DBG
    FastIoDispatch->FastIoRead                  = Ext2FastIoRead;
    FastIoDispatch->FastIoWrite                 = Ext2FastIoWrite;
#else
    FastIoDispatch->FastIoRead                  = FsRtlCopyRead;
    FastIoDispatch->FastIoWrite                 = FsRtlCopyWrite;
#endif
    FastIoDispatch->FastIoQueryBasicInfo        = Ext2FastIoQueryBasicInfo;
    FastIoDispatch->FastIoQueryStandardInfo     = Ext2FastIoQueryStandardInfo;
    FastIoDispatch->FastIoLock                  = Ext2FastIoLock;
    FastIoDispatch->FastIoUnlockSingle          = Ext2FastIoUnlockSingle;
    FastIoDispatch->FastIoUnlockAll             = Ext2FastIoUnlockAll;
    FastIoDispatch->FastIoUnlockAllByKey        = Ext2FastIoUnlockAllByKey;
    FastIoDispatch->FastIoQueryNetworkOpenInfo  = Ext2FastIoQueryNetworkOpenInfo;

    DriverObject->FastIoDispatch = FastIoDispatch;

    switch ( MmQuerySystemSize() ) {

    case MmSmallSystem:

        Ext2Global->MaxDepth = 16;
        break;

    case MmMediumSystem:

        Ext2Global->MaxDepth = 64;
        break;

    case MmLargeSystem:

        Ext2Global->MaxDepth = 256;
        break;
    }

    //
    // Initialize the Cache Manager callbacks
    //
    
    CacheManagerCallbacks = &(Ext2Global->CacheManagerCallbacks);
    CacheManagerCallbacks->AcquireForLazyWrite  = Ext2AcquireForLazyWrite;
    CacheManagerCallbacks->ReleaseFromLazyWrite = Ext2ReleaseFromLazyWrite;
    CacheManagerCallbacks->AcquireForReadAhead  = Ext2AcquireForReadAhead;
    CacheManagerCallbacks->ReleaseFromReadAhead = Ext2ReleaseFromReadAhead;

    Ext2Global->CacheManagerNoOpCallbacks.AcquireForLazyWrite  = Ext2NoOpAcquire;
    Ext2Global->CacheManagerNoOpCallbacks.ReleaseFromLazyWrite = Ext2NoOpRelease;
    Ext2Global->CacheManagerNoOpCallbacks.AcquireForReadAhead  = Ext2NoOpAcquire;
    Ext2Global->CacheManagerNoOpCallbacks.ReleaseFromReadAhead = Ext2NoOpRelease;


    //
    // Initialize the global data
    //

    InitializeListHead(&(Ext2Global->VcbList));
    ExInitializeResourceLite(&(Ext2Global->Resource));
    ExInitializeResourceLite(&(Ext2Global->CountResource));
    ExInitializeResourceLite(&(Ext2Global->LAResource));

    ExInitializeNPagedLookasideList( &(Ext2Global->Ext2IrpContextLookasideList),
                                     NULL,
                                     NULL,
                                     0,
                                     sizeof(EXT2_IRP_CONTEXT),
                                     '2TXE',
                                     0 );

    ExInitializeNPagedLookasideList( &(Ext2Global->Ext2FcbLookasideList),
                                     NULL,
                                     NULL,
                                     0,
                                     sizeof(EXT2_FCB),
                                     '2TXE',
                                     0 );

    ExInitializeNPagedLookasideList( &(Ext2Global->Ext2CcbLookasideList),
                                     NULL,
                                     NULL,
                                     0,
                                     sizeof(EXT2_CCB),
                                     '2TXE',
                                     0 );

    ExInitializePagedLookasideList( &(Ext2Global->Ext2McbLookasideList),
                                     NULL,
                                     NULL,
                                     0,
                                     sizeof(EXT2_MCB),
                                     '2TXE',
                                     0 );

#if EXT2_UNLOAD
    RtlInitUnicodeString(&DosDeviceName, DOS_DEVICE_NAME);
    IoCreateSymbolicLink(&DosDeviceName, &DeviceName);
#endif

#if DBG
    ProcessNameOffset = Ext2GetProcessNameOffset();
#endif

    Ext2LoadAllNls();

    PAGE_TABLE = load_nls(Ext2Global->CodePage.AnsiName);

    IoRegisterFileSystem(DeviceObject);
    
    return Status;
}
